
COLOR_BLACK = (0,0,0)
COLOR_WHITE = (255,255,255)

# RGB
## RED
COLOR_RED = (255, 0, 0)
COLOR_RED_DARK  = (100, 0, 0)
COLOR_RED_LIGHT = (255, 100, 100)
## GREEN
COLOR_GREEN = (0, 255, 0)
COLOR_GREEN_DARK  = (0, 100, 0)
COLOR_GREEN_LIGHT = (100, 255, 100)
COLOR_GREEN_GRASS = (80, 240, 100) 
## BLUE
COLOR_BLUE = (0, 0, 255)
COLOR_BLUE_DARK  = (0, 0, 100)
COLOR_BLUE_LIGHT = (100, 100, 255)

# CYM
## CYAN
COLOR_CYAN = (0, 255, 255)
COLOR_CYAN_DARK  = (0, 100, 100)
COLOR_CYAN_LIGHT = (100, 255, 255)
## YELLOW
COLOR_YELLOW = (255, 255, 0)
COLOR_YELLOW_DARK  = (100, 100, 0)
COLOR_YELLOW_LIGHT = (255, 255, 100)
## MAGENTA
COLOR_MAGENTA = (255, 0, 255)
COLOR_MAGENTA_DARK  = (100, 0, 100)
COLOR_MAGENTA_LIGHT = (255, 100, 255)